Require Export realbasis.
Module REAL_FACTS(Nums : NUMS).

Module Nums_ext := NUMS_EXT(Nums).
Export Nums_ext.

Module definitions := EXACT_REAL_DEFINITIONS(Nums).
Export definitions.

Open Scope R_scope.

(* The relation linking Stream to standard definition of real numbers in Coq*)
CoInductive Srepresents : Stream digit -> R -> Prop :=
  | rep : forall s r k, 
    Srepresents s r ->
    (-1 <= r <= 1)%R -> 
    (- b < IDZ k < b)%Z ->
    Srepresents (k::s) ((IDR k +r)/ Z2R b)%R.

Inductive Srep_head : Stream digit -> nat -> R -> Prop :=
  | rep_head_O : forall s, Srep_head s 0%nat 0%R
  | rep_head_S : forall x1 x n vx , Srep_head x n vx ->
    (- b < IDZ x1 < b)%Z ->
    Srep_head (x1::x) (S n) ((IDR x1 + vx)/Z2R b).

Inductive Srep_tail : Stream digit -> nat -> R -> Prop :=
  | rep_tail_O : forall x vx, Srepresents x vx -> Srep_tail x 0%nat vx
  | rep_tail_S : forall x1 x n vx , Srep_tail x n vx ->
    Srep_tail (x1::x) (S n) vx.

(* The relation linking (exponant, stream) to standard real numbers in Coq *)
Inductive represents : (nat * Stream digit) -> R -> Prop :=
  | rep_R : forall e s r, Srepresents s r ->
    represents (e, s) (Z2R b^e *r)%R.

Ltac str_unfold term :=
  rewrite (unfold_Stream term).

Ltac fold_decomp :=
  match goal with
    |- context [ match ?a with h::t => h::t end] =>
      rewrite <- (unfold_Stream a)
  end.

Ltac inversion_to_variable x :=
  match goal with
  | id : Srepresents x _ |- _ => idtac
  | id : Srepresents (Cons ?d x) _ |- context [Cons ?d x] => idtac
  | id : Srepresents (Cons ?d1 x) _ |-  _ => inversion id
  | id : Srepresents (Cons ?d1 (Cons ?d2 x)) _ |- _ =>
    inversion id; inversion_to_variable x
  end.

Ltac represents_tac :=
  match goal with 
    id : Srepresents ?x ?r1
    |- represents (?e, ?x) ?r2 =>
      replace r2 with (Z2R b^ e*r1)%R; 
	[constructor; exact id| idtac]
  end.

Ltac rep_constr := 
  match goal with 
    | id : Srepresents ?x ?r1 
      |- Srepresents (?k::?x) ?r2 => 
	replace r2 with ((IDR k + r1)/Z2R b)%R
    | id : Srepresents ?x ?r1 
      |- Srepresents (?k1::?k2::?x) ?r2 => 
	replace r2 with ((IDR k1 + (IDR k2 + r1)/Z2R b)/Z2R b)%R
  end.

Ltac compare_tac H := 
  match goal with 
    |- context [Dcompare ?a ?b] => 
      generalize (compare_IDZ a b);
	elim (Dcompare a b); simpl_IDZ; intros H
  end.

Theorem compare_IDZ_elim : 
  forall (A : Type) (x1 x2 : digit) (y1 y2 y3: A) ,
    let y := 
      match (x1 ?= x2)%D with 
	| Eq => y1
	| Lt => y2
	| Gt => y3
      end in  
      (IDZ x1 = IDZ x2 /\ y = y1)%Z \/
      (IDZ x1 < IDZ x2 /\ y = y2)%Z \/ 
      (IDZ x2 < IDZ x1 /\ y = y3)%Z. 
Proof.
intros A x1 x2 y1 y2 y3.
elim (Ztrichotomy (IDZ x1) (IDZ x2)); intros G1;
  [right; left | elim G1; intros G2; [left|right; right]];
  (split; [omega| ]); compare_tac G3;
    auto; absurd (IDZ x1 = IDZ x2); omega.
Qed.

Theorem compare_Eq_IDZ : 
   forall (A : Type) (x1 x2 : digit) (y1 y2 : A) ,
     (IDZ x1 = IDZ x2 /\ 
       match (x1 ?= x2)%D with 
	 | Eq => y1
	 | _ => y2
       end = y1) \/
     (IDZ x1 <> IDZ x2 /\ 
       match (x1 ?= x2)%D with 
	 | Eq => y1
	 | _ => y2
       end = y2).  
Proof.
intros A x1 x2 y1 y2.
elim (Z_eq_dec (IDZ x1) (IDZ x2)); intros G1;[ left | right];
  (split; [omega|]);
  (elim (compare_IDZ_elim _ x1 x2 y1 y2 y2);
    [| intros G2; elim G2; clear G2];
    intros [G2 G3]; (assumption || absurd (IDZ x1 = IDZ x2);  omega)).
Qed.

Theorem compare_Lt_IDZ : 
   forall (A : Type) (x1 x2 : digit) (y1 y2 : A) ,
     (IDZ x1 < IDZ x2 /\ 
       match (x1 ?= x2)%D with 
	 | Lt => y1
	 | _ => y2
       end = y1)%Z \/
     (IDZ x2 <= IDZ x1 /\ 
       match (x1 ?= x2)%D with 
	 | Lt => y1
	 | _ => y2
       end = y2)%Z.  
Proof.
intros A x1 x2 y1 y2.
elim (Z_lt_dec (IDZ x1) (IDZ x2)); intros G1;[ left | right];
  (split; [omega|]);
  (elim (compare_IDZ_elim _ x1 x2 y2 y1 y2);
    [| intros G2; elim G2; clear G2];
    intros [G2 G3]; (assumption || absurd (IDZ x1 = IDZ x2);  omega)).
Qed.

Theorem compare_Gt_IDZ : 
   forall (A : Type) (x1 x2 : digit) (y1 y2 : A) ,
     (IDZ x2 < IDZ x1 /\ 
       match (x1 ?= x2)%D with 
	 | Gt => y1
	 | _ => y2
       end = y1)%Z \/
     (IDZ x1 <= IDZ x2 /\ 
       match (x1 ?= x2)%D with 
	 | Gt => y1
	 | _ => y2
       end = y2)%Z.  
Proof.
intros A x1 x2 y1 y2.
elim (Z_gt_dec (IDZ x1) (IDZ x2)); intros G1;[ left | right];
  (split; [omega|]);
  (elim (compare_IDZ_elim _ x1 x2 y2 y2 y1);
    [| intros G2; elim G2; clear G2];
    intros [G2 G3]; (assumption || absurd (IDZ x1 = IDZ x2);  omega)).
Qed.

Ltac compare_eq_tac H :=
  match goal with 
    | |- context [match Dcompare ?x1 ?x2 with | Eq => ?y1 | Lt => ?y2 | Gt => ?y2 end] =>
      let H' := fresh "H" in 
	elim (compare_Eq_IDZ _ x1 x2 y1 y2); simpl_IDZ;
	  intros [H H']; rewrite H'; clear H'
  end.

Ltac compare_lt_tac H :=
  match goal with 
    | |- context [match Dcompare ?x1 ?x2 with | Eq => ?y2 | Lt => ?y1 | Gt => ?y2 end] =>
      let H' := fresh "H" in 
	elim (compare_Lt_IDZ _ x1 x2 y1 y2); simpl_IDZ;
	  intros [H H']; rewrite H'; clear H'
  end.

Ltac compare_gt_tac H :=
  match goal with 
    | |- context [match Dcompare ?x1 ?x2 with | Eq => ?y2 | Lt => ?y2 | Gt => ?y1 end] =>
      let H' := fresh "H" in 
	elim (compare_Gt_IDZ _ x1 x2 y1 y2); simpl_IDZ;
	  intros [H H']; rewrite H'; clear H'
  end.

(*TODO remove it when Scompare will be used everywhere*)
Ltac Zbool H := 
  match goal with 
    |- context [Zle_bool ?a ?b] => 
      generalize (Zle_cases a b);
	elim (Zle_bool_total_tf a b); 
	  intros H; rewrite H; clear H; simpl_IDZ; intro H
  end.

(*
Ltac Ddiv_eucl_tac q r H1 H2 H3 H4:= 
match goal with 
  |- context [ Ddiv_eucl ?a ?b] => 
    let H := fresh "H" in 
    assert (H : ( IDZ b >0)%Z);[ simpl_IDZ; ZR0compat|
      generalize (Z_div_mod_eq (IDZ a) (IDZ b) H);
	generalize (Z_mod_lt (IDZ a) (IDZ b) H);
	  generalize (div_IDZ a b);
	    generalize (mod_IDZ a b);
	      caseEq (Ddiv_eucl a b);
	      intros q r _ H1 H2;
		rewrite <- H1; rewrite <- H2; intros H3 H4]
end.
*)

(*Apply tactic replace_f that replace the real argument of Srepresents such that 
  we can apply the constructor of Srepresents and then the cofix hypothesis
   f_correct*)
Ltac apply_cofix replace_f f f_correct :=
  replace_f f; 
  [ constructor; 
      [ apply f_correct; clear f_correct | clear f_correct | clear f_correct ] 
    | clear f_correct; try field_IDR].

Ltac replace_for_repeat_digit repeat_digit:=
  match goal with 
    |- Srepresents (?k::repeat_digit ?k) ?v => 
      replace v with ((IDR k + (IDR k / (Z2R b-1)))/Z2R b)%R
  end. 

Theorem repeat_digit_correct :
  forall k, 
    (-   b < IDZ k <  b)%Z ->
    Srepresents (repeat_digit k) (IDR k/ (Z2R b-1))%R.
Proof.
cofix.
assert (Hbpos:= base_positive).
intros k Hk.
str_unfold (repeat_digit k); simpl.
apply_cofix replace_for_repeat_digit repeat_digit repeat_digit_correct;
 assumption || idtac.
apply Rle_div_1; [R_to_Z|].
split; IDR_to_Z.
Qed.

Theorem rep_Szero : 
  Srepresents Szero 0%R.
Proof.
generalize base_positive; intros Hbpos. 
unfold Szero. 
replace 0%R with ((IDR D0)/ (Z2R b-1))%R by field_IDR.
apply repeat_digit_correct; simpl_IDZ; omega.
Qed. 

Theorem rep_Sone : 
  Srepresents Sone 1%R.
Proof.
generalize base_positive; intros Hbpos.
unfold Sone. 
replace 1%R with ((IDR Dmax)/ (Z2R b-1))%R by field_IDR.
apply repeat_digit_correct; simpl_IDZ; omega.
Qed. 

Theorem rep_Sminus_one : 
  Srepresents Sminus_one (-1)%R.
Proof.
generalize base_positive; intros Hbpos. 
unfold Sminus_one. 
replace (-1)%R with ((IDR (-Dmax)%D)/ (Z2R b-1))%R by field_IDR.
apply repeat_digit_correct; simpl_IDZ; omega.
Qed.

Ltac const_rep :=
  match goal with 
    |- Srepresents Sone ?r => 
      replace r with 1%R; [apply rep_Sone| idtac]
    | |- Srepresents Sminus_one ?r => 
      replace r with (-1)%R; [apply rep_Sminus_one| idtac]
    | |- Srepresents Szero ?r => 
      replace r with 0%R; [apply rep_Szero| idtac]
  end.

Theorem rep_zero : 
  represents zero 0%R.
Proof.
unfold zero. 
replace 0%R with (Z2R b^0%nat*0)%R by ring.
constructor.
const_rep; auto.
Qed. 

Theorem rep_one : 
  represents one 1%R.
Proof.
unfold one. 
replace 1%R with (Z2R b^0%nat*1)%R.
constructor; const_rep; auto.
ring.
Qed. 

Theorem I_interval :
  forall x v k,
    Srepresents (k::x) v ->
    ((IDR k - 1)/Z2R b <= v <= (IDR k + 1)/Z2R b)%R.
Proof.
intros x v k Hrep.
inversion Hrep; subst.
assert (G1 : (0 < Z2R b)%R) by R_to_Z.
split; apply Rdiv_le_compat_r; try assumption; dec_tac; fourier.
Qed.

Theorem I_interval' :
  forall x v k,
    Srepresents (k::x) v ->
    (IDR k - 1 <= v*Z2R b <= IDR k + 1)%R.
Proof.
intros x v k Hrep.
inversion Hrep; subst.
assert (0< Z2R b)%R by R_to_Z.
rewrite Rmult_comm.
rewrite Rmult_div_simpl; [ | ZR0compat].
dec_tac; split; fourier.
Qed.

Theorem Srepresents_interval : 
  forall x vx, 
    Srepresents x vx -> 
    (-1 <= vx <= 1)%R.
Proof.
intros x vx Hrep.
inversion Hrep; subst.
apply Rle_div_1; [R_to_Z|].
assert (-Z2R b+1<= IDR k <= Z2R b -1) by (split; IDR_to_Z).
dec_tac; split; fourier.
Qed.

Theorem Srep_head_inter : 
  forall k x vx,
    Srep_head x k vx -> 
    Rabs vx <= 1 - 1/ Z2R b^k.
Proof.   
intros k; induction k; intros x vx Hrepx; simpl; 
inversion Hrepx; subst.

rewrite Rabs_R0; right; field.

rewrite Rabs_div; [| R_to_Z].
rewrite (Rabs_right (Z2R b)); [| ZR0compat].
apply Rle_div_r;[R_to_Z| ].
replace ((1 - 1 / (Z2R b * Z2R b ^ k)) * Z2R b) with 
  ( (Z2R b -1) + (1 - 1/Z2R b ^k)) ;[| field_Z; ZR0compat].
apply Rle_trans with (1:=  (Rabs_triang (IDR x1) vx0)).
apply Rplus_le_compat.
rabs_elim; split; IDR_to_Z.
apply IHk with x0; assumption.
Qed.

Theorem Srep_head_inter_1 : 
  forall k x vx,
    Srep_head x k vx -> 
    Rabs vx <= 1.
Proof.   
assert (Hbpos:= base_positive).
intros k x vx Hrep.
replace 1 with (1 + 0) by ring.
apply Rle_trans with (1:= (Srep_head_inter _ _ _ Hrep)).
unfold Rminus; apply Rplus_le_compat_l.
ZR0compat.
Qed.

Theorem Srep_tail_inter : 
  forall k x vx,
    Srep_tail x k vx -> 
    Rabs vx <= 1.
Proof.   
intros k; induction k; intros x vx Hrepx; simpl; 
inversion Hrepx; subst.
rabs_elim; apply (Srepresents_interval _ _ H).

apply IHk with x0; assumption.
Qed.

Theorem Srep_head_tail : 
  forall k x vx_head vx_tail, 
    Srep_head x k vx_head -> 
    Srep_tail x k vx_tail ->
    Srepresents x (vx_head+vx_tail/Z2R b ^k).
Proof.
intros k; induction k; simpl; intros x vx_head vx_tail Hrepx_h Hrepx_t.
inversion Hrepx_h; inversion Hrepx_t.
replace (0 + vx_tail /1) with vx_tail by field; assumption.

inversion Hrepx_h; inversion Hrepx_t; subst.
injection H5;intros; subst.
assert (R1 :  (IDR x1 + vx) / Z2R b + vx_tail/ (Z2R b * Z2R b ^ k) =  
  (IDR x1 + (vx+ vx_tail/Z2R b^k)) / Z2R b); [field_Z; ZR0compat |].
rewrite R1; constructor; [| | assumption].
apply IHk; assumption.
rabs_intros.
apply Rle_trans with (1:= Rabs_triang vx (vx_tail / Z2R b ^ k)).
replace 1 with ((1 - 1/Z2R b^ k) + 1/Z2R b^ k) by (unfold Rdiv; ring).
apply Rplus_le_compat.
apply Srep_head_inter with x0; assumption.
rewrite Rabs_div; [| ZR0compat].
rewrite (Rabs_right (Z2R b^k)); [| ZR0compat].
apply Rdiv_le_compat_r; [ZR0compat|].
apply Srep_tail_inter with k x0; assumption. 
Qed.

Theorem Srep_tail_str_tail : 
  forall k x vx, 
    Srepresents (Str_nth_tl k x) vx <-> 
    Srep_tail x k vx.
Proof.
intros k; induction k; simpl; intros x vx.
split; intros Hrep.
constructor; assumption.
inversion Hrep; assumption.

destruct x as [x1 x'].
simpl; rewrite IHk.
split; intros Hrep.
constructor; assumption.
inversion Hrep; assumption.
Qed.

Lemma lim_conv_low:
forall (un:nat->R) (a l:R),
(forall n, a<=un n)->
Un_cv un l ->
a<=l.
Proof.
intros un t l H H0.
cut (~(l<t)).
intros H1.
apply Rnot_gt_le; auto.
unfold not; intros.
unfold Un_cv in H0.
assert (t-l>0).
 fourier.
assert (Hcon: exists N, forall n, (n>=N)%nat ->R_dist (un n) l <t-l).
apply (H0 (t-l) H2).
elim Hcon; intros N Hlim.
assert (Hdist:R_dist (un N) l<t-l).
apply (Hlim N).
omega.
unfold R_dist in Hdist.
assert (Ha: t<=un N).
apply (H N).
rewrite Rabs_right in Hdist; [|fourier].
assert (un N<t); fourier.
Qed.

Lemma lim_seq_monot:
forall (xn yn:nat->R) (x y:R), 
(forall n, xn n<= yn n)->
(Un_cv xn x)->
(Un_cv yn y)->
x<=y.
intros.
assert (0<=y-x).
assert (forall n, 0<=yn n- xn n).
intros.
elim (H n);intros; fourier.
assert (Un_cv (fun i:nat => yn i - xn i) (y - x)).
apply CV_minus; assumption.
apply lim_conv_low with (fun  i:nat => yn i - xn i);
assumption.
fourier.
Qed.

Lemma eq_aux:
 forall a  xn, (forall n, Rabs ( a ) <= xn n) ->
  Un_cv xn 0 -> a = 0.
Proof.
intros a xn Habs Hcv. 
assert (H1: Rabs a <= 0).
apply (lim_seq_monot (fun n => Rabs a) xn).
assumption.
intros e he; exists 0%nat; intros n Hn; unfold R_dist.
replace (Rabs a - Rabs a) with 0 by ring.
rewrite Rabs_R0; assumption.
assumption.
assert (Rabs a = 0).
apply Rle_antisym ; [assumption|apply Rabs_pos].
elim (Req_dec a 0); intros H0; [assumption| elim (Rabs_no_R0 _ H0); assumption].
Qed.

Theorem Srep_eq : 
  forall k x vx1 vx2,
    Srepresents x vx1 -> 
    Srepresents x vx2 ->
    Z2R b^k * Rabs (vx1 - vx2) <= 2.
Proof.
intros k; induction k; intros x vx1 vx2 Hrepx1 Hrepx2; simpl.
assert (H1 := Srepresents_interval _ _ Hrepx1).
assert (H2 := Srepresents_interval _ _ Hrepx2).
ring_simplify.
rabs_elim.
dec_tac; split; fourier.
inversion Hrepx1; inversion Hrepx2; subst.
injection H7; intros; subst.

replace (Z2R b * Z2R b ^ k * Rabs ((IDR k0 + r) / Z2R b - (IDR k0 + r0) / Z2R b)) 
  with ( Z2R b ^ k * Rabs (r-r0)).
apply IHk with s; assumption.

replace ((IDR k0 + r) / Z2R b - (IDR k0 + r0) / Z2R b) with 
(( r - r0) / Z2R b) by (unfold Rdiv; ring).
rewrite Rabs_div; [| R_to_Z].
rewrite (Rabs_right (Z2R b)); [| ZR0compat].
field_Z.
Qed.

Lemma eq_real_of_stream:
  forall x vx1 vx2,
    Srepresents x vx1 -> 
    Srepresents x vx2 ->
    vx1 = vx2.
Proof.
assert (Hbpos:= base_positive).
intros x vx1 vx2 Hrepx1 Hrepx2.
assert (H:forall k,   Z2R b^k * Rabs (vx1 - vx2) <= 2).
intros k; apply Srep_eq with x ; assumption. 
assert (vx1 - vx2 = 0).
apply eq_aux with (fun n => 2/ Z2R b ^n).
intros n.
apply Rle_div_l.
ZR0compat.
rewrite Rmult_comm; apply H.
intros e He.
assert (He2: 0 < e/2).
fourier.
elim ( archim_lt2 _  He2); intros N HN.
exists N; intros n Hn; unfold R_dist.
apply Rle_lt_trans with (2* 1/2^N). 
rewrite Rminus_0_r.
rewrite Rabs_right; [|ZR0compat].
apply Rle_trans with (2*1 / 2 ^ n).
rewrite Rmult_1_r.
apply Rdiv_mult_compat; ZR0compat.
apply Rmult_le_compat_l; ZR0compat.
apply pow_incr;split; R_to_Z.
rewrite Rmult_1_r.
apply Rdiv_mult_compat; ZR0compat.
apply Rmult_le_compat_l; ZR0compat.
apply Rle_pow; ZR0compat.
apply Rmult_lt_reg_l with (/2).
fourier.
replace (  / 2 * (2 * 1 / 2 ^ N)) with (1/2^N).
rewrite Rmult_comm; assumption.
field; ZR0compat.
apply Rplus_eq_reg_l with (-vx2).
rewrite Rplus_comm; unfold Rminus in H0; rewrite H0.
ring.
Qed.

Theorem Srep_Szero_eq_0 : 
  forall x,
    Srepresents Szero x -> 
    0 = x.
Proof.
intros x Hrepx; apply eq_real_of_stream with Szero;
 [apply rep_Szero|assumption].
Qed.

Theorem bounds_aux_correct : 
  forall n acc x vx,
    Srepresents x vx -> 
    -1 <= (IBR acc + vx)*Z2R b^n - IBR (bounds_aux n acc x) <= 1.
Proof.
intros n; induction n;
intros acc x vx Hrepx;simpl.
replace ((IBR acc+  vx) * 1 - IBR acc)%R with vx by ring.
apply (Srepresents_interval _ _ Hrepx).

destruct x as [k x']; inversion Hrepx; subst.
replace ((IBR acc + (IDR k + r) / Z2R b) * (Z2R b * Z2R b ^ n)) with 
  ((IBR (acc * Bbase + big_of_digit k)%B + r) * Z2R b ^ n) by field_IBR.
apply IHn; assumption.
Qed.

Theorem bounds_correct : 
  forall n x vx,
    Srepresents x vx -> 
    -1 <= vx*Z2R b^n - IBR (bounds n x) <= 1.
Proof.
intros n x vx Hrepx.
replace (vx * Z2R b ^ n - IBR (bounds n x)) with 
  ((IBR 0%B + vx) * Z2R b ^ n - IBR (bounds n x)) by (simpl_IBR; ring).
unfold bounds.
apply bounds_aux_correct; assumption.
Qed.

Ltac replace_for_multb multb  :=
  match goal with 
    id : Srepresents ?x ?vx |- Srepresents (?k::multb (?d::?x)) ?v => 
      replace v with ((IDR k + Z2R b * ((IDR d +vx)/Z2R b))/Z2R b)%R
  end.  

Theorem multb_correct : 
  forall x r, 
    Srepresents x r ->
    (-1 <= Z2R b*r <= 1)%R ->
    Srepresents (multb x) (Z2R b*r)%R.
Proof.
cofix.
intros x r Hrep Hbound.

assert (U1 : 0 < Z2R b).
  generalize base_positive; intros; R_to_Z.

str_unfold (multb x).
destruct x as [k1 x']; inversion Hrep; subst.
replace  (Z2R b * ((IDR k1 + r0) / Z2R b)) with 
  (IDR k1 + r0) in * by field_IDR.
simpl.
fold_decomp.
compare_tac G1.

replace (IDR k1) with 1%R in * by (simpl_IDR; rewrite G1; auto).
destruct x' as [k2 x'']; inversion_to_variable x''; subst.
compare_tac G2.

replace (IDR k2) with 0%R in * by (simpl_IDR; rewrite G2; auto).
rewrite Rplus_0_l in *.
assert (U3 : (-1 <= Z2R b * ((IDR 1%D + r) / Z2R b) <= 1)%R).
  clear multb_correct.
  replace (Z2R b * ((IDR 1%D + r) / Z2R b))%R with (1 + r)%R by field_IDR.
  split; [dec_tac; fourier|].
  assert (r <= 0);[|fourier].
  replace r with (Z2R b * (r / Z2R b)) by field_Z.
  replace 0 with (Z2R b * 0) by ring.
  apply Rmult_le_compat_l; [fourier|].
  generalize_all (r / Z2R b) r1; dec_tac; fourier.
replace (1 + r / Z2R b) with 
  ((IDR Dmax + Z2R b * (IDR 1%D + r)/ Z2R b) / Z2R b) by field_IDR.
apply_cofix replace_for_multb multb multb_correct; 
  assumption ||  idtac.
constructor; (assumption || simpl_IDZ; omega).
simpl_IDZ; omega.

replace (1 + (IDR k2 + r) / Z2R b) with 
  ((IDR (k2 + Dmax + 1)%D +r)/Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.

clear multb_correct.
const_rep.
apply Rle_antisym; [| intuition].
pattern 1 at 1; replace 1 with (1 + 0) by ring.
apply Rplus_le_compat_l.
apply Rdiv_le_0_compat; [|auto].
assert (1 <= IDR k2) by IDR_to_Z; dec_tac; fourier.

compare_tac G2.

replace (IDR k1) with (-1)%R in * by (simpl_IDR; rewrite G2; auto).
destruct x' as [k2 x'']; inversion_to_variable x''; subst.
compare_tac G3.

replace (IDR k2) with 0%R in * by (simpl_IDR; rewrite G3; auto).
rewrite Rplus_0_l in *.
assert (U3 : (-1 <= Z2R b * ((IDR (-(1))%D + r) / Z2R b) <= 1)%R).
  clear multb_correct.
  replace (Z2R b * ((IDR (-(1))%D + r) / Z2R b))%R with (-1 + r)%R by field_IDR.
  split; [|dec_tac; fourier].
  assert (0 <= r);[|fourier].
  replace r with (Z2R b * (r / Z2R b)) by field_Z.
  replace 0 with (Z2R b * 0) by ring.
  apply Rmult_le_compat_l; [fourier|].
  generalize_all (r / Z2R b) r1; dec_tac; fourier.
replace (-1 + r / Z2R b) with 
  ((IDR (-Dmax)%D + Z2R b * (IDR (-(1))%D + r)/ Z2R b) / Z2R b) by field_IDR.
apply_cofix replace_for_multb multb multb_correct; 
  assumption ||  idtac.
constructor; (assumption || simpl_IDZ; omega).
simpl_IDZ; omega.

clear multb_correct.
const_rep.
apply Rle_antisym; [ intuition|].
pattern (-1) at 2; replace (-1) with (-1 + 0/Z2R b) by (unfold Rdiv; ring).
apply Rplus_le_compat_l.
apply Rdiv_le_compat_r;[auto|].
assert (IDR k2 <= -1) by IDR_to_Z; dec_tac; fourier.

replace (-1 + (IDR k2 + r) / Z2R b) with 
  ((IDR (k2 - Dmax - 1)%D +r)/Z2R b) by field_IDR.
constructor; assumption || simpl_IDZ; omega.

clear multb_correct.
const_rep.
apply Rle_antisym; [ intuition|].
assert (IDR k1 <= -2) by IDR_to_Z; dec_tac; fourier.

clear multb_correct.
replace (IDR k1) with 0 by IDR_to_Z.
rewrite Rplus_0_l; assumption.

clear multb_correct.
const_rep.
apply Rle_antisym; [| intuition].
assert (2 <= IDR k1) by IDR_to_Z; dec_tac; fourier.
Qed.

Theorem Sopp_correct : 
  forall x vx, 
    Srepresents x vx ->
    Srepresents (Sopp x) (-vx)%R.
Proof.
cofix.
intros x vx Hrepx.
str_unfold (Sopp x).
simpl.
destruct x as [k x'].
inversion Hrepx; subst.
replace (- ((IDR k + r) / Z2R b))%R with ((IDR (-k)%D+-r)/Z2R b)%R by field_IDR.
constructor.
apply Sopp_correct; assumption.
clear Sopp_correct.
dec_tac; split; fourier.
simpl_IDZ; omega.
Qed.

Theorem opp_correct : 
  forall x vx, 
    represents x vx ->
    represents (opp x) (-vx)%R.
Proof.
intros x vx Hrepx.
inversion Hrepx; subst.
unfold opp.
replace (- ( Z2R b ^e* r)) with (Z2R b^e * (-r)) by ring.
constructor.
apply Sopp_correct; assumption.
Qed.
  
Ltac rat_to_str_tac :=
  match goal with 
    |- Srepresents (?k::rat_to_stream_pos ?a ?c) ?r =>
      replace r with ((IDR  k + (IBR a/ IBR c))/ Z2R b)%R
    | |- Srepresents (?k::rat_to_stream_neg ?a ?c) ?r =>
      replace r with ((IDR  k%D + (-(IBR a/ IBR c)))/ Z2R b)%R
end.

Theorem rat_to_stream_pos_correct :
  forall a c,
    (0 <= IBZ a < IBZ c)%Z  ->
    (-1<=IBR a/ IBR c<=1)%R ->
    Srepresents (rat_to_stream_pos  a c) (IBR a/ IBR c)%R.
Proof.
cofix.
intros a c Hacpos Hdiv.
str_unfold (rat_to_stream_pos a c).
simpl.
fold_decomp.
generalize (div_IBZ (a * Bbase)%B c).
generalize (mod_IBZ (a * Bbase)%B c).
caseEq (Bdiv_eucl (a * Bbase)%B c).
intros k r Hdiveq Hr Hk.
generalize base_positive; intros G2.
assert (G3 : (0<= IBZ r < IBZ c)%Z).
rewrite Hr; apply Z_mod_lt; omega.
assert (G4 : ( -1 <= IBR r / IBR c <= 1)%R).
apply Rle_div_1.
IBR_to_Z.
split; IBR_to_Z.

rat_to_str_tac.
constructor; try apply rat_to_stream_pos_correct; 
  clear rat_to_stream_pos_correct; try assumption.

simpl_IDZ.
assert (0<= IBZ k < b)%Z; [| omega].
rewrite Hk.
simpl_IBZ.
apply Zdiv_mult_bound; (assumption|| omega).

simpl_IDR.
assert (G5: (IBZ c > 0)%Z) by omega.
generalize (Z_div_mod_eq (IBZ (a * Bbase)%B) (IBZ c) G5).
rewrite <- Hk.
rewrite <- Hr.
simpl_IBZ.
intros G6.
field_simplify_eq; [idtac| split; [ IBR_to_Z| IDR_to_Z]].
IBR_to_Z.
rewrite Zmult_comm.
rewrite <- G6; ring. 
Qed.

Theorem rat_to_stream_neg_correct :
  forall a c,
    (0 <= IBZ a < IBZ c)%Z  ->
    (-1<=IBR a/ IBR c<=1)%R ->
    Srepresents (rat_to_stream_neg a c) (-(IBR a/ IBR c))%R.
Proof.
cofix.
intros a c Hacpos Hdiv.
str_unfold (rat_to_stream_neg a c).
simpl.
fold_decomp.
generalize (div_IBZ (a * Bbase)%B c).
generalize (mod_IBZ (a * Bbase)%B c).
caseEq (Bdiv_eucl (a * Bbase)%B c).
intros k r Hdiveq Hr Hk.
generalize base_positive; intros G2.
assert (G3 : (0<= IBZ r < IBZ c)%Z).
rewrite Hr; apply Z_mod_lt; omega.
assert (G4 : ( -1 <= IBR r / IBR c <= 1)%R).
apply Rle_div_1.
IBR_to_Z.
split; IBR_to_Z.
rat_to_str_tac.
constructor; try apply rat_to_stream_neg_correct; 
  clear rat_to_stream_neg_correct; try assumption.
generalize_all (IBR r / IBR c) x.
intuition; fourier.
simpl_IDZ; simpl_IBZ.
assert (0<= IBZ k < b)%Z; [| omega].
rewrite Hk.
simpl_IBZ.
apply Zdiv_mult_bound; (assumption|| omega).
simpl_IDR.
assert (G5: (IBZ c > 0)%Z).
omega.
generalize (Z_div_mod_eq (IBZ (a * Bbase)%B) (IBZ c) G5).
rewrite <- Hk.
rewrite <- Hr.
simpl_IBZ.
intros G6.
field_simplify_eq; [idtac| split; [ IBR_to_Z| IDR_to_Z]].
IBR_to_Z.
replace (- b * IBZ a)%Z with (-(IBZ a* b))%Z by ring.
rewrite G6; ring. 
Qed.

Theorem Srat_to_stream_correct :
  forall a c, 
    (0<> IBZ c)%Z  ->
    (-1<=IBR a/ IBR c<=1)%R ->
    Srepresents (Srat_to_stream a c) (IBR a/ IBR c)%R.
Proof.
intros a c Hcneq0 [Hdiv1 Hdiv2].
unfold Srat_to_stream.
simpl_IBZ.
assert (U1 : (IBR a/ IBR c = -(IBR (-a)%B/IBR c))%R).
simpl_IBR; unfold Rdiv; ring.
assert (U2 : (IBR a/ IBR c = -(IBR a/IBR (-c)%B))%R).
simpl_IBR; field_IBR.
assert (U3 : (IBR a/ IBR c = IBR (-a)%B/IBR (-c)%B)%R).
simpl_IBR; field_IBR.

Zbool G1; Zbool G2; Zbool G3; try (const_rep; apply Rle_antisym; try assumption). 

apply Rle_div_l;IBR_to_Z.

apply rat_to_stream_pos_correct; simpl_IBZ; (omega || intuition).

apply Rle_div_r; IBR_to_Z.

rewrite U1.
apply rat_to_stream_neg_correct; (simpl_IBZ; omega || simpl_IBR).
rewrite Ropp_div_distr_l_reverse.
fold_IBR; intuition; generalize Hdiv1; generalize (IBR a/ IBR c);intros; fourier.

rewrite U3.
apply Rle_div_l; IBR_to_Z.

rewrite U3.
apply rat_to_stream_pos_correct; simpl_IBZ; 
  (omega || rewrite <- U3; intuition).

rewrite U2.
apply Ropp_le_contravar.
apply Rle_div_l; IBR_to_Z.

rewrite U2.
apply rat_to_stream_neg_correct; (simpl_IBZ; omega || simpl_IBR).
rewrite Ropp_div_distr_r_reverse.
fold_IBR; generalize Hdiv1 Hdiv2; generalize (IBR a/ IBR c); intros; intuition; fourier.
IBR_to_Z.
Qed.

Theorem Zpower_Z_of_nat_S : 
  forall z n, 
    (z ^ Z_of_nat (S n) = z * z ^ Z_of_nat n)%Z.
Proof.
intros z n.
replace (S n) with (n+1)%nat by ring.
rewrite inj_plus.
rewrite Zpower_exp; ZR0compat.
change (Z_of_nat 1) with 1%Z; ring.
Qed.

Theorem Blog_aux_correct : 
  (forall x n y l y' l',
      IBZ x <= IBZ y * b ^ (Z_of_nat n) ->
      Blog_aux x y l n = (y', l') -> 
      (l<= l')%nat /\ IBZ x <= IBZ y' /\ IBZ y' = IBZ y * b ^ Z_of_nat (l'-l))%Z.
Proof.
intros x n; induction n; intros y l y' l' Hxy; simpl Blog_aux.
injection 1; intros; subst.
replace (l'-l')%nat with O by omega; rewrite Zmult_1_r in *; auto.

simpl_IBZ.
Zbool G1.
injection 1; intros; subst.
replace (l'-l')%nat with O by omega; rewrite Zmult_1_r;auto.

intros Heq.

assert (Hxy' :  (IBZ x <= IBZ (Bbase*y)%B * b ^ Z_of_nat n)%Z).
  apply Zle_trans with (1:= Hxy).
  apply Zeq_le.
  simpl_IBZ; rewrite Zpower_Z_of_nat_S; ring.

generalize (IHn _ _ _ _ Hxy' Heq); intros [U1 [U2 U3]].
split; [intuition|].
split; [intuition|].
rewrite U3; simpl_IBZ.
replace (l' - l)%nat with (S (l' - S l))%nat by omega.
rewrite Zpower_Z_of_nat_S; ring.
Qed.

Theorem Zpower_log_le : 
  forall b q, 
    (1 < b ->
      q+1 <= b ^q)%Z.
Proof.
intros b' q Hbpos.
case q; [simpl; omega
  |
  | intros p; simpl Zpower;
    assert (Zneg p <0)%Z by apply Zlt_neg_0; 
      omega].

intros p.
rewrite <- (Z_of_nat_Zabs_nat (Zpos p)); [|auto with zarith].
generalize (Zabs_nat (Zpos p)); clear p.
intros n; induction n.
simpl; omega.

rewrite Zpower_Z_of_nat_S.
apply Zle_trans with (b' * (Z_of_nat n + 1))%Z.
  replace (Z_of_nat (S n) + 1 )%Z with (1*Z_of_nat n +2)%Z; [|
    replace (S n) with (n + 1)%nat by ring; 
      rewrite inj_plus; ring].
  replace ( b' * (Z_of_nat n + 1))%Z with ( b' * Z_of_nat n + b')%Z by ring.
  apply Zplus_le_compat; [|omega].
  apply Zmult_le_compat_r; omega.

  apply Zmult_le_compat_l; omega.
Qed.

Theorem Blog_correct : 
  forall x y l y',
    (0 <= IBZ x)%Z ->
    (0 < IBZ y)%Z ->
    Blog_aux (x+1)%B y 0 (Zabs_nat (IBZ x / IBZ y)%Z) = (y', l) ->
    (IBZ x < IBZ y')%Z /\ IBZ y' = (IBZ y * b ^ Z_of_nat l)%Z.
Proof.
intros x y l y' Hx Hy.
set (q:= (IBZ x / IBZ y)%Z).
set (r:= Zmod (IBZ x) (IBZ y)).
assert (U1 : (IBZ x = IBZ y * q +r)%Z).   
  rewrite (Z_div_mod_eq (IBZ x) (IBZ y));[|omega].
  unfold q, r; ring.

intros Heq.

cut (IBZ (x+1)%B <= IBZ y * b ^ Z_of_nat (Zabs_nat q))%Z.
intros Hxy.

generalize (Blog_aux_correct _ (Zabs_nat q) y 0 y' l Hxy Heq).
simpl_IBZ; rewrite <- minus_n_O; intuition.

simpl_IBZ; rewrite U1.
apply Zle_trans with (IBZ y * (q +1))%Z.
  replace (IBZ y * (q +1))%Z with (IBZ y * q+ IBZ y)%Z by ring.
  rewrite <- Zplus_assoc.
  apply Zplus_le_compat_l.
  assert (0<= r < IBZ y)%Z; [| omega].
  unfold r; apply Z_mod_lt; omega.

  apply Zmult_le_compat_l; [|omega].
  generalize base_positive; intros Hbpos.
  rewrite Z_of_nat_Zabs_nat; [apply Zpower_log_le; omega|] .
  assert (q >= 0)%Z; [|omega].
  unfold q; apply Z_div_ge0; omega.
Qed.

Theorem rat_to_stream_correct : 
  forall x y,
    (0 <> IBZ y)%Z -> 
    represents (rat_to_stream x y) (IBR x/ IBR y).
Proof.
intros x y Hynot0.
pose (Hbpos := base_positive).

unfold rat_to_stream.
simpl_IBZ.
Zbool G1; Zbool G2; try Zbool G3.

generalize (div_IBZ x y).
generalize (mod_IBZ x y).
caseEq (Bdiv_eucl x y).
intros q r Heq Hmod Hdiv.
let_simpl; intros y' l Heq'.
assert (U1 : (0 <= IBZ x)%Z) by omega.
assert (U2 : (0 < IBZ y)%Z) by omega.
rewrite Hdiv in Heq'.
generalize (Blog_correct x y l y' U1 U2 Heq').
intros [U3 U4].
assert (U5: (0 < IBR y')).
  unfold IBR; rewrite U4; ZR0compat.
replace (IBR x / IBR y) with (Z2R b ^ l * (IBR x / IBR y'));
  [| unfold IBR, IDR; rewrite U4;  simpl_Z2R; 
    field_Z; ZR0compat].
constructor.
apply Srat_to_stream_correct; [ omega |  ].
apply Rle_div_1 with (1:= U5).
split; IBR_to_Z.

generalize (div_IBZ x (-y)%B).
generalize (mod_IBZ x (-y)%B).
caseEq (Bdiv_eucl x (-y)%B).
intros q r Heq Hmod Hdiv.
let_simpl; intros y' l Heq'.
assert (U1 : (0 <= IBZ x)%Z) by omega.
assert (U2 : (0 < IBZ (-y)%B)%Z) by (simpl_IBZ; omega).
rewrite Hdiv in Heq'.
generalize (Blog_correct x (-y)%B l y' U1 U2 Heq').
intros [U3 U4].
assert (U5: (0 < IBR y')).
  unfold IBR; rewrite U4; ZR0compat.
replace (IBR x / IBR y) with (Z2R b ^ l * (IBR x / IBR (-y')%B));
  [| simpl_IBR; unfold IBR, IDR; rewrite U4; simpl_IBZ;
    simpl_Z2R; field_Z; ZR0compat].
constructor.
apply Srat_to_stream_correct ; [simpl_IBZ; omega |].
replace (IBR x / IBR (- y')%B) with (- IBR x / IBR y') by field_IBR.
apply Rle_div_1 ; [generalize U2; simpl_IBZ; intros; IBR_to_Z|].
split; IBR_to_Z.

replace (IBR x) with 0 by IBR_to_Z.
replace (0 / IBR y) with 0 by (unfold Rdiv; ring).
apply rep_zero; omega.

generalize (div_IBZ (-x)%B y).
generalize (mod_IBZ (-x)%B y).
caseEq (Bdiv_eucl (-x)%B y).
intros q r Heq Hmod Hdiv.
let_simpl; intros y' l Heq'.
assert (U1 : (0 <= IBZ (-x)%B)%Z) by (simpl_IBZ; omega).
assert (U2 : (0 < IBZ y)%Z) by (omega).
rewrite Hdiv in Heq'.
generalize (Blog_correct  (-x)%B y l y' U1 U2 Heq').
simpl_IBZ; intros [U3 U4].
assert (U5: (0 < IBR y')).
  unfold IBR; rewrite U4; ZR0compat.
replace (IBR x / IBR y) with (Z2R b ^ l * (IBR x / IBR y'));
  [| simpl_IBR; unfold IBR, IDR; rewrite U4; 
    simpl_Z2R; field_Z; ZR0compat].
constructor.
apply Srat_to_stream_correct ; [simpl_IBZ; omega |].
apply Rle_div_1 ; [generalize U2; simpl_IBZ; intros; IBR_to_Z|].
split; IBR_to_Z.

generalize (div_IBZ (-x)%B (-y)%B).
generalize (mod_IBZ (-x)%B (-y)%B).
caseEq (Bdiv_eucl (-x)%B (-y)%B).
intros q r Heq Hmod Hdiv.
let_simpl; intros y' l Heq'.
assert (U1 : (0 <= IBZ (-x)%B)%Z) by (simpl_IBZ; omega).
assert (U2 : (0 < IBZ (-y)%B)%Z) by (simpl_IBZ; omega).
rewrite Hdiv in Heq'.
generalize (Blog_correct (-x)%B (-y)%B l y' U1 U2 Heq').
simpl_IBZ; intros [U3 U4].
assert (U5: (0 < IBR y')).
  unfold IBR; rewrite U4; ZR0compat.
replace (IBR x / IBR y) with (Z2R b ^ l * (IBR x / IBR (-y')%B));
  [| simpl_IBR; unfold IBR, IDR; rewrite U4; simpl_IBZ;
    simpl_Z2R; field_Z; ZR0compat].
constructor.
apply Srat_to_stream_correct ; [simpl_IBZ; omega |].
replace (IBR x / IBR (- y')%B) with (- IBR x / IBR y') by field_IBR.
apply Rle_div_1 ; [generalize U2; simpl_IBZ; intros; IBR_to_Z|].
split; IBR_to_Z.
Qed.

Theorem rep_R_plus_div : 
  forall e1 e2 x r, 
    represents (e1+e2, x)%nat r ->
    represents (e1, x) (r / Z2R b^e2)%R.
Proof.
intros e1 e2 x r Hrep.
inversion Hrep; subst.
represents_tac.
rewrite pow_add; try omega.
field.
unfold IDR; ZR0compat.
generalize base_positive; omega.
Qed.

Theorem rep_R_div_plus : 
  forall e1 e2 x r, 
    represents (e1, x) (r / Z2R b^ e2) ->
    represents (e1+e2, x)%nat r.
Proof.
intros e1 e2 x r Hrep.
inversion Hrep; subst.
represents_tac.
rewrite pow_add.
match goal with 
  |- ?a*?b*?c = ?d => 
    replace (a*b*c) with ((a*c)*b) by ring
end.
rewrite H2; field; unfold IDR; ZR0compat.
generalize base_positive; omega.
Qed.

Theorem str_del_zero_correct : 
  forall e x r, 
    represents (e, x) r ->
    represents (str_del_zero e x) r.
Proof.
intros e; induction e; intros x r Hrep.
simpl; tauto.
destruct x as [k x'].
simpl.
compare_eq_tac G1; [| assumption].
apply IHe.
inversion Hrep; subst.
inversion_to_variable x'; subst.
represents_tac.
replace (IDR k) with 0%R by IDR_to_Z.
replace (S e) with (e+1)%nat by ring.
rewrite pow_add.
rewrite pow_1.
field_IDR.
Qed.

Theorem Rdiv_1_r : 
  forall r, 
    r/1=r.
Proof.
intros; field.
Qed.

Theorem str_add_zero_correct : 
  forall e x r,
    Srepresents x r ->
    Srepresents (str_add_zero e x) (r/ Z2R b^ e).
Proof.
intros e.
induction e.
intros x r Hrep.
simpl.
rewrite Rdiv_1_r.
assumption.

intros x r Hrep.
pose (Hbpos := base_positive).
simpl str_add_zero.
replace (S e) with (1+e)%nat by ring.
rewrite pow_add.
rewrite pow_1.
rewrite <- Rdiv_div_mult.
apply IHe.
rep_constr.
constructor; assumption || simpl_IDZ; omega || idtac.
apply Srepresents_interval with (1:=Hrep). 
simpl_IDR; unfold Rdiv; ring.
IDR_to_Z.
ZR0compat.
Qed.

End REAL_FACTS.
